import os
import copy
import logging
import datetime
from itertools import product
from collections import defaultdict
from typing import Callable, Dict, List, Tuple, Union

import csv
import h5py
import numpy as np
import pandas as pd

from ml4h.metrics import weighted_crossentropy
from ml4h.normalizer import Standardize, ZeroMeanStd1
from ml4h.TensorMap import TensorMap, str2date, Interpretation, make_range_validator, decompress_data, TimeSeriesOrder
from ml4h.defines import ECG_REST_AMP_LEADS, PARTNERS_DATE_FORMAT, STOP_CHAR, PARTNERS_DATETIME_FORMAT, CARDIAC_SURGERY_DATE_FORMAT

YEAR_DAYS = 365.26
INCIDENCE_CSV = '/media/erisone_snf13/lc_outcomes.csv'
CARDIAC_SURGERY_OUTCOMES_CSV = '/data/sts-data/mgh-preop-ecg-outcome-labels.csv'
PARTNERS_PREFIX = 'partners_ecg_rest'


def _hd5_filename_to_mrn_int(filename: str) -> int:
    return int(os.path.basename(filename).split('.')[0])


def _get_ecg_dates(tm, hd5):
    if not hasattr(_get_ecg_dates, 'mrn_lookup'):
        _get_ecg_dates.mrn_lookup = dict()
    mrn = _hd5_filename_to_mrn_int(hd5.filename)
    if mrn in _get_ecg_dates.mrn_lookup:
        return _get_ecg_dates.mrn_lookup[mrn]

    dates = list(hd5[tm.path_prefix])
    if tm.time_series_lookup is not None:
        start, end = tm.time_series_lookup[mrn]
        dates = [date for date in dates if start < date < end]
    if tm.time_series_order == TimeSeriesOrder.NEWEST:
        dates.sort()
    elif tm.time_series_order == TimeSeriesOrder.OLDEST:
        dates.sort(reverse=True)
    elif tm.time_series_order == TimeSeriesOrder.RANDOM:
        np.random.shuffle(dates)
    else:
        raise NotImplementedError(f'Unknown option "{tm.time_series_order}" passed for which tensors to use in multi tensor HD5')
    start_idx = tm.time_series_limit if tm.time_series_limit is not None else 1
    dates = dates[-start_idx:]  # If num_tensors is 0, get all tensors
    dates.sort(reverse=True)
    _get_ecg_dates.mrn_lookup[mrn] = dates
    return dates


# Date formatting
def _partners_str2date(d) -> datetime.date:
    return datetime.datetime.strptime(d, PARTNERS_DATE_FORMAT).date()


def validator_no_empty(tm: TensorMap, tensor: np.ndarray, hd5: h5py.File):
    if any(tensor == ''):
        raise ValueError(f'TensorMap {tm.name} failed empty string check.')


def validator_no_negative(tm: TensorMap, tensor: np.ndarray, hd5: h5py.File):
    if any(tensor < 0):
        raise ValueError(f'TensorMap {tm.name} failed non-negative check')


def validator_not_all_zero(tm: TensorMap, tensor: np.ndarray, hd5: h5py.File):
    if np.count_nonzero(tensor) == 0:
        raise ValueError(f'TensorMap {tm.name} failed all-zero check')


def _is_dynamic_shape(tm: TensorMap, num_ecgs: int) -> Tuple[bool, Tuple[int, ...]]:
    if tm.shape[0] is None:
        return True, (num_ecgs,) + tm.shape[1:]
    return False, tm.shape


def _make_hd5_path(tm, ecg_date, value_key):
    return f'{tm.path_prefix}/{ecg_date}/{value_key}'


def _resample_voltage(voltage, desired_samples):
    if len(voltage) == desired_samples:
        return voltage
    elif len(voltage) == 2500 and desired_samples == 5000:
        x = np.arange(2500)
        x_interp = np.linspace(0, 2500, 5000)
        return np.interp(x_interp, x, voltage)
    elif len(voltage) == 5000 and desired_samples == 2500:
        return voltage[::2]
    else:
        raise ValueError(f'Voltage length {len(voltage)} is not desired {desired_samples} and re-sampling method is unknown.')


def _resample_voltage_with_rate(voltage, desired_samples, rate, desired_rate):
    if len(voltage) == desired_samples and rate == desired_rate:
        return voltage
    elif desired_samples / len(voltage) == 2 and desired_rate / rate == 2:
        x = np.arange(len(voltage))
        x_interp = np.linspace(0, len(voltage), desired_samples)
        return np.interp(x_interp, x, voltage)
    elif desired_samples / len(voltage) == 0.5 and desired_rate / rate == 0.5:
        return voltage[::2]
    elif desired_samples / len(voltage) == 2 and desired_rate == rate:
        return np.pad(voltage, (0, len(voltage)))
    elif desired_samples / len(voltage) == 0.5 and desired_rate == rate:
        return voltage[:len(voltage)//2]
    else:
        raise ValueError(f'Voltage length {len(voltage)} is not desired {desired_samples} with desired rate {desired_rate} and rate {rate}.')


def make_voltage(exact_length = False):
    def get_voltage_from_file(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        voltage_length = shape[1] if dynamic else shape[0]
        tensor = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            for cm in tm.channel_map:
                try:
                    path = _make_hd5_path(tm, ecg_date, cm)
                    voltage = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
                    if exact_length:
                        assert len(voltage) == voltage_length
                    voltage = _resample_voltage(voltage, voltage_length)
                    slices = (i, ..., tm.channel_map[cm]) if dynamic else (..., tm.channel_map[cm])
                    tensor[slices] = voltage
                except (KeyError, AssertionError, ValueError):
                    logging.debug(f'Could not get voltage for lead {cm} with {voltage_length} samples in {hd5.filename}')
        return tensor
    return get_voltage_from_file


def voltage_stat(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.zeros(shape, dtype=np.float32)
    for i, ecg_date in enumerate(ecg_dates):
        try:
            slices = lambda stat: (i, tm.channel_map[stat]) if dynamic else (tm.channel_map[stat],)
            path = lambda lead: _make_hd5_path(tm, ecg_date, lead)
            voltages = np.array([decompress_data(data_compressed=hd5[path(lead)][()], dtype='int16') for lead in ECG_REST_AMP_LEADS])
            tensor[slices('mean')] = np.mean(voltages)
            tensor[slices('std')] = np.std(voltages)
            tensor[slices('min')] = np.min(voltages)
            tensor[slices('max')] = np.max(voltages)
            tensor[slices('median')] = np.median(voltages)
        except KeyError:
            logging.warning(f'Could not get voltage stats for ECG at {hd5.filename}')
    return tensor


partners_ecg_voltage_stats = TensorMap(
    'partners_ecg_voltage_stats',
    shape=(None, 5),
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=voltage_stat,
    channel_map={'mean': 0, 'std': 1, 'min': 2, 'max': 3, 'median': 4},
    time_series_limit=0,
)


def make_voltage_attr(volt_attr: str = ""):
    def get_voltage_attr_from_file(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            for cm in tm.channel_map:
                try:
                    path = _make_hd5_path(tm, ecg_date, cm)
                    slices = (i, tm.channel_map[cm]) if dynamic else (tm.channel_map[cm],)
                    tensor[slices] = hd5[path].attrs[volt_attr]
                except KeyError:
                    pass
        return tensor
    return get_voltage_attr_from_file


voltage_len = TensorMap(
    "voltage_len",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_voltage_attr(volt_attr="len"),
    shape=(None, 12),
    channel_map=ECG_REST_AMP_LEADS,
    time_series_limit=0,
)


def make_partners_ecg_label(keys: Union[str, List[str]] = "read_md_clean", dict_of_list: Dict = dict(), not_found_key: str = "unspecified"):
    if type(keys) == str:
        keys = [keys]

    def get_partners_ecg_label(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        label_array = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            found = False
            for channel, idx in sorted(tm.channel_map.items(), key=lambda cm: cm[1]):
                if channel not in dict_of_list:
                    continue
                for key in keys:
                    path = _make_hd5_path(tm, ecg_date, key)
                    if path not in hd5:
                        continue
                    read = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
                    for string in dict_of_list[channel]:
                        if string not in read:
                            continue
                        slices = (i, idx) if dynamic else (idx,)
                        label_array[slices] = 1
                        found = True
                        break
                    if found:
                        break
                if found:
                    break
            if not found:
                slices = (i, tm.channel_map[not_found_key]) if dynamic else (tm.channel_map[not_found_key],)
                label_array[slices] = 1
        return label_array
    return get_partners_ecg_label


def partners_ecg_datetime(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.full(shape , '', dtype=f'<U19')
    for i, ecg_date in enumerate(ecg_dates):
        tensor[i] = ecg_date
    return tensor


partners_ecg_datetime = TensorMap(
    "partners_ecg_datetime",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=partners_ecg_datetime,
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


def make_partners_ecg_tensor(key: str, fill: float = 0, channel_prefix: str = '', channel_unknown: str = 'other'):
    def get_partners_ecg_tensor(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        if tm.interpretation == Interpretation.LANGUAGE:
            tensor = np.full(shape, '', dtype=object)
        elif tm.interpretation == Interpretation.CONTINUOUS:
            tensor = np.zeros(shape, dtype=np.float32) if fill == 0 else np.full(shape, fill, dtype=np.float32)
        elif tm.interpretation == Interpretation.CATEGORICAL:
            tensor = np.zeros(shape, dtype=float)
        else:
            raise NotImplementedError(f'unsupported interpretation for partners tmaps: {tm.interpretation}')

        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date, key)
            try:
                data = decompress_data(data_compressed=hd5[path][()], dtype='str')
                if tm.interpretation == Interpretation.CATEGORICAL:
                    matched = False
                    data = f'{channel_prefix}{data}'
                    for cm in tm.channel_map:
                        if data.lower() == cm.lower():
                            slices = (i, tm.channel_map[cm]) if dynamic else (tm.channel_map[cm],)
                            tensor[slices] = 1.0
                            matched = True
                            break
                    if not matched:
                        slices = (i, tm.channel_map[channel_unknown]) if dynamic else (tm.channel_map[channel_unknown],)
                        tensor[slices] = 1.0
                else:
                    tensor[i] = data
            except (KeyError, ValueError):
                logging.debug(f'Could not obtain tensor {tm.name} from ECG on {ecg_date} in {hd5.filename}')

        if tm.interpretation == Interpretation.LANGUAGE:
            tensor = tensor.astype(str)
        return tensor
    return get_partners_ecg_tensor


def make_partners_language_tensor(key: str):
    def language_tensor(tm, hd5, dependents={}):
        words = str(decompress_data(data_compressed=hd5[key][()], dtype=hd5[key].attrs['dtype']))
        tensor = np.zeros(tm.shape, dtype=np.float32)
        for i, c in enumerate(words):
            if i >= tm.shape[0]:
                logging.debug(f'Text {words} is longer than {tm.name} can store in shape:{tm.shape}, truncating...')
                break
            tensor[i, tm.channel_map[c]] = 1.0
        tensor[min(tm.shape[0]-1, i+1), tm.channel_map[STOP_CHAR]] = 1.0
        return tensor
    return language_tensor


partners_ecg_read_md = TensorMap(
    "partners_ecg_read_md",
    #annotation_units=128,
    #channel_map=PARTNERS_CHAR_2_IDX,
    interpretation=Interpretation.LANGUAGE,
    #shape=(512, len(PARTNERS_CHAR_2_IDX)),
    path_prefix=PARTNERS_PREFIX,
    #tensor_from_file=make_partners_language_tensor(key="read_md_clean"),
    tensor_from_file=make_partners_ecg_tensor(key="read_md_clean"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_read_pc = TensorMap(
    "partners_ecg_read_pc",
    #annotation_units=128,
    #channel_map=PARTNERS_CHAR_2_IDX,
    interpretation=Interpretation.LANGUAGE,
    #tensor_from_file=make_partners_language_tensor(key="read_pc_clean"),
    #shape=(512, len(PARTNERS_CHAR_2_IDX)),
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="read_pc_clean"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_patientid = TensorMap(
    "partners_ecg_patientid",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="patientid"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


def validator_clean_mrn(tm: TensorMap, tensor: np.ndarray, hd5: h5py.File):
    int(tensor)


partners_ecg_patientid_clean = TensorMap(
    "partners_ecg_patientid_clean",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="patientid_clean"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_clean_mrn,
)


partners_ecg_firstname = TensorMap(
    "partners_ecg_firstname",
    #channel_map=PARTNERS_CHAR_2_IDX,
    interpretation=Interpretation.LANGUAGE,
    #tensor_from_file=make_partners_language_tensor(key="patientfirstname"),
    #shape=(512, len(PARTNERS_CHAR_2_IDX)),
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="patientfirstname"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_lastname = TensorMap(
    "partners_ecg_lastname",
    #channel_map=PARTNERS_CHAR_2_IDX,
    interpretation=Interpretation.LANGUAGE,
    #tensor_from_file=make_partners_language_tensor(key="patientlastname"),
    #shape=(512, len(PARTNERS_CHAR_2_IDX)),
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="patientlastname"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_sex = TensorMap(
    "partners_ecg_sex",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="gender"),
    channel_map={'female': 0, 'male': 1},
    time_series_limit=0,
    validator=validator_not_all_zero,
)

partners_ecg_date = TensorMap(
    "partners_ecg_date",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="acquisitiondate"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_time = TensorMap(
    "partners_ecg_time",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="acquisitiontime"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_sitename = TensorMap(
    "partners_ecg_sitename",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="sitename"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_location = TensorMap(
    "partners_ecg_location",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="location"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


partners_ecg_dob = TensorMap(
    "partners_ecg_dob",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="dateofbirth"),
    shape=(None, 1),
    time_series_limit=0,
    validator=validator_no_empty,
)


def make_sampling_frequency_from_file(lead: str = "I", duration: int = 10, channel_prefix: str = "_", channel_unknown: str = "other", fill: int = -1):
    def sampling_frequency_from_file(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        if tm.interpretation == Interpretation.CATEGORICAL:
            tensor = np.zeros(shape, dtype=np.float32)
        else:
            tensor = np.full(shape, fill, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date, lead)
            lead_length = hd5[path].attrs["len"]
            sampling_frequency = lead_length / duration
            try:
                if tm.interpretation == Interpretation.CATEGORICAL:
                    matched = False
                    sampling_frequency = f'{channel_prefix}{sampling_frequency}'
                    for cm in tm.channel_map:
                        if sampling_frequency.lower() == cm.lower():
                            slices = (i, tm.channel_map[cm]) if dynamic else (tm.channel_map[cm],)
                            tensor[slices] = 1.0
                            matched = True
                            break
                    if not matched:
                        slices = (i, tm.channel_map[channel_unknown]) if dynamic else (tm.channel_map[channel_unknown],)
                        tensor[slices] = 1.0
                else:
                    tensor[i] = sampling_frequency
            except (KeyError, ValueError):
                logging.debug(f'Could not calculate sampling frequency from ECG on {ecg_date} in {hd5.filename}')
        return tensor
    return sampling_frequency_from_file


partners_ecg_sampling_frequency = TensorMap(
    "partners_ecg_sampling_frequency",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_sampling_frequency_from_file(),
    channel_map={'_250': 0, '_500': 1, 'other': 2},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_sampling_frequency_pc = TensorMap(
    "partners_ecg_sampling_frequency_pc",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="ecgsamplebase_pc", channel_prefix='_'),
    channel_map={'_0': 0, '_250': 1, '_500': 2, 'other': 3},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_sampling_frequency_md = TensorMap(
    "partners_ecg_sampling_frequency_md",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="ecgsamplebase_md", channel_prefix='_'),
    channel_map={'_0': 0, '_250': 1, '_500': 2, 'other': 3},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_sampling_frequency_wv = TensorMap(
    "partners_ecg_sampling_frequency_wv",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="waveform_samplebase", channel_prefix='_'),
    channel_map={'_0': 0, '_240': 1, '_250': 2, '_500': 3, 'other': 4},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_sampling_frequency_continuous = TensorMap(
    "partners_ecg_sampling_frequency_continuous",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_sampling_frequency_from_file(),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_sampling_frequency_pc_continuous = TensorMap(
    "partners_ecg_sampling_frequency_pc_continuous",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="ecgsamplebase_pc", fill=-1),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_sampling_frequency_md_continuous = TensorMap(
    "partners_ecg_sampling_frequency_md_continuous",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="ecgsamplebase_md", fill=-1),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_sampling_frequency_wv_continuous = TensorMap(
    "partners_ecg_sampling_frequency_wv_continuous",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="waveform_samplebase", fill=-1),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_time_resolution = TensorMap(
    "partners_ecg_time_resolution",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="intervalmeasurementtimeresolution", channel_prefix='_'),
    channel_map={'_25': 0, '_50': 1, '_100': 2, 'other': 3},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_amplitude_resolution = TensorMap(
    "partners_ecg_amplitude_resolution",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="intervalmeasurementamplituderesolution", channel_prefix='_'),
    channel_map={'_10': 0, '_20': 1, '_40': 2, 'other': 3},
    time_series_limit=0,
    validator=validator_not_all_zero,
)


partners_ecg_measurement_filter = TensorMap(
    "partners_ecg_measurement_filter",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="intervalmeasurementfilter", channel_prefix='_'),
    time_series_limit=0,
    channel_map={'_None': 0, '_40': 1, '_80': 2, 'other': 3},
    validator=validator_not_all_zero,
)


partners_ecg_high_pass_filter = TensorMap(
    "partners_ecg_high_pass_filter",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="waveform_highpassfilter", fill=-1),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_low_pass_filter = TensorMap(
    "partners_ecg_low_pass_filter",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="waveform_lowpassfilter", fill=-1),
    time_series_limit=0,
    shape=(None, 1),
    validator=validator_no_negative,
)


partners_ecg_ac_filter = TensorMap(
    "partners_ecg_ac_filter",
    interpretation=Interpretation.CATEGORICAL,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="waveform_acfilter", channel_prefix='_'),
    time_series_limit=0,
    channel_map={'_None': 0, '_50': 1, '_60': 2, 'other': 3},
    validator=validator_not_all_zero,
)


partners_ecg_rate_pc = TensorMap(
    "partners_ecg_rate_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="ventricularrate_pc"),
    shape=(None, 1),
    normalization=Standardize(mean=59.3, std=10.6),
    time_series_limit=0,
    validator=make_range_validator(10, 200),
)


partners_ecg_rate_md = TensorMap(
    "partners_ecg_rate_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="ventricularrate_md"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(10, 200),
)


partners_ecg_qrs_pc = TensorMap(
    "partners_ecg_qrs_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qrsduration_pc"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(20, 400),
)


partners_ecg_qrs_md = TensorMap(
    "partners_ecg_qrs_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qrsduration_md"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(20, 400),
)


partners_ecg_pr_pc = TensorMap(
    "partners_ecg_pr_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="printerval_pc"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(50, 500),
)


partners_ecg_pr_md = TensorMap(
    "partners_ecg_pr_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="printerval_md"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(50, 500),
)


partners_ecg_qt_pc = TensorMap(
    "partners_ecg_qt_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qtinterval_pc"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(100, 800),
)


partners_ecg_qt_md = TensorMap(
    "partners_ecg_qt_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qtinterval_md"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(100, 800),
)


partners_ecg_qtc_pc = TensorMap(
    "partners_ecg_qtc_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qtcorrected_pc"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(100, 800),
)


partners_ecg_qtc_md = TensorMap(
    "partners_ecg_qtc_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="qtcorrected_md"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(100, 800),
)


partners_ecg_paxis_pc = TensorMap(
    "partners_ecg_paxis_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="paxis_pc", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_paxis_md = TensorMap(
    "partners_ecg_paxis_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="paxis_md", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_raxis_pc = TensorMap(
    "partners_ecg_raxis_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="raxis_pc", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_raxis_md = TensorMap(
    "partners_ecg_raxis_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="raxis_md", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_taxis_pc = TensorMap(
    "partners_ecg_taxis_pc",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="taxis_pc", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_taxis_md = TensorMap(
    "partners_ecg_taxis_md",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="taxis_md", fill=999),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(-180, 180),
)


partners_ecg_measuredamplitudepeak_r = TensorMap(
    "partners_ecg_measuredamplitudepeak_r",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="measuredamplitudepeak_IE_R", fill=np.nan),
    shape=(None, 12),
    time_series_limit=0,
)


partners_ecg_acquisitiondevice = TensorMap(
    "partners_ecg_acquisitiondevice",
    interpretation=Interpretation.LANGUAGE,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=make_partners_ecg_tensor(key="acquisitiondevice", fill=999),
    shape=(None, 1),
    time_series_limit=0,
)


partners_ecg_weight_lbs = TensorMap(
    "partners_ecg_weight_lbs",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    loss='log_cosh',
    tensor_from_file=make_partners_ecg_tensor(key="weightlbs"),
    shape=(None, 1),
    time_series_limit=0,
    validator=make_range_validator(100, 800),
)


def _partners_ecg_age_from_hd5(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.zeros(shape, dtype=float)
    for i, ecg_date in enumerate(ecg_dates):
        if i >= shape[0]:
            break
        path = lambda key: _make_hd5_path(tm, ecg_date, key)
        try:
            birthday = decompress_data(data_compressed=hd5[path('dateofbirth')][()], dtype='str')
            acquisition = decompress_data(data_compressed=hd5[path('acquisitiondate')][()], dtype='str')
            delta = _partners_str2date(acquisition) - _partners_str2date(birthday)
            years = delta.days / YEAR_DAYS
            tensor[i] = years
        except KeyError:
            try:
                tensor[i] = decompress_data(data_compressed=hd5[path('patientage')][()], dtype='str')
            except KeyError:
                logging.debug(f'Could not get patient date of birth or age from ECG on {ecg_date} in {hd5.filename}')
    return tensor


partners_ecg_age = TensorMap('partners_ecg_age', path_prefix=PARTNERS_PREFIX, loss='log_cosh', tensor_from_file=_partners_ecg_age_from_hd5, shape=(None, 1), time_series_limit=0)


def partners_ecg_acquisition_year(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.zeros(shape, dtype=int)
    for i, ecg_date in enumerate(ecg_dates):
        path = _make_hd5_path(tm, ecg_date, 'acquisitiondate')
        try:
            acquisition = decompress_data(data_compressed=hd5[path][()], dtype='str')
            tensor[i] = _partners_str2date(acquisition).year
        except KeyError:
            pass
    return tensor


partners_ecg_acquisition_year = TensorMap('partners_ecg_acquisition_year', path_prefix=PARTNERS_PREFIX, loss='log_cosh',  tensor_from_file=partners_ecg_acquisition_year, shape=(None, 1), time_series_limit=0)


def partners_bmi(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.zeros(shape, dtype=float)
    for i, ecg_date in enumerate(ecg_dates):
        path = lambda key: _make_hd5_path(tm, ecg_date, key)
        try:
            weight_lbs = decompress_data(data_compressed=hd5[path('weightlbs')][()], dtype='str')
            weight_kg = 0.453592 * float(weight_lbs)
            height_in = decompress_data(data_compressed=hd5[path('heightin')][()], dtype='str')
            height_m = 0.0254 * float(height_in)
            bmi = weight_kg / (height_m*height_m)
            logging.info(f' Height was {height_in} weight: {weight_lbs} bmi is {bmi}')
            tensor[i] = bmi
        except KeyError:
            pass
    return tensor


partners_ecg_bmi = TensorMap('partners_ecg_bmi', path_prefix=PARTNERS_PREFIX, channel_map={'bmi': 0}, tensor_from_file=partners_bmi, time_series_limit=0)


def partners_channel_string(hd5_key, race_synonyms={}, unspecified_key=None):
    def tensor_from_string(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            path = _make_hd5_path(tm, ecg_date,hd5_key)
            found = False
            try:
                hd5_string = decompress_data(data_compressed=hd5[path][()], dtype='str')
                for key in tm.channel_map:
                    slices = (i, tm.channel_map[key]) if dynamic else (tm.channel_map[key],)
                    if hd5_string.lower() == key.lower():
                        tensor[slices] = 1.0
                        found = True
                        break
                    if key in race_synonyms:
                        for synonym in race_synonyms[key]:
                            if hd5_string.lower() == synonym.lower():
                                tensor[slices] = 1.0
                                found = True
                            if found: break
                        if found: break
            except KeyError:
                pass
            if not found:
                if unspecified_key is None:
                    # TODO Do we want to try to continue to get tensors for other ECGs in HD5?
                    raise ValueError(f'No channel keys found in {hd5_string} for {tm.name} with channel map {tm.channel_map}.')
                slices = (i, tm.channel_map[unspecified_key]) if dynamic else (tm.channel_map[unspecified_key],)
                tensor[slices] = 1.0
        return tensor
    return tensor_from_string


race_synonyms = {'asian': ['oriental'], 'hispanic': ['latino'], 'white': ['caucasian']}
partners_ecg_race = TensorMap(
    'partners_ecg_race', interpretation=Interpretation.CATEGORICAL, path_prefix=PARTNERS_PREFIX, channel_map={'asian': 0, 'black': 1, 'hispanic': 2, 'white': 3, 'unknown': 4},
    tensor_from_file=partners_channel_string('race', race_synonyms), time_series_limit=0,
)


def _partners_adult(hd5_key, minimum_age=18):
    def tensor_from_string(tm, hd5, dependents={}):
        ecg_dates = _get_ecg_dates(tm, hd5)
        dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
        tensor = np.zeros(shape, dtype=np.float32)
        for i, ecg_date in enumerate(ecg_dates):
            path = lambda key: _make_hd5_path(tm, ecg_date, key)
            birthday = decompress_data(data_compressed=hd5[path('dateofbirth')][()], dtype='str')
            acquisition = decompress_data(data_compressed=hd5[path('acquisitiondate')][()], dtype='str')
            delta = _partners_str2date(acquisition) - _partners_str2date(birthday)
            years = delta.days / YEAR_DAYS
            if years < minimum_age:
                raise ValueError(f'ECG taken on patient below age cutoff.')
            hd5_string = decompress_data(data_compressed=hd5[path(hd5_key)][()], dtype=hd5[path(hd5_key)].attrs['dtype'])
            found = False
            for key in tm.channel_map:
                if hd5_string.lower() == key.lower():
                    slices = (i, tm.channel_map[key]) if dynamic else (tm.channel_map[key],)
                    tensor[slices] = 1.0
                    found = True
                    break
            if not found:
                # TODO Do we want to try to continue to get tensors for other ECGs in HD5?
                raise ValueError(f'No channel keys found in {hd5_string} for {tm.name} with channel map {tm.channel_map}.')
        return tensor
    return tensor_from_string


partners_adult_sex = TensorMap(
    'adult_sex', interpretation=Interpretation.CATEGORICAL, path_prefix=PARTNERS_PREFIX, channel_map={'female': 0, 'male': 1},
    tensor_from_file=_partners_adult('gender'), time_series_limit=0,
)


def voltage_zeros(tm, hd5, dependents={}):
    ecg_dates = _get_ecg_dates(tm, hd5)
    dynamic, shape = _is_dynamic_shape(tm, len(ecg_dates))
    tensor = np.zeros(shape, dtype=np.float32)
    for i, ecg_date in enumerate(ecg_dates):
        for cm in tm.channel_map:
            path = _make_hd5_path(tm, ecg_date, cm)
            voltage = decompress_data(data_compressed=hd5[path][()], dtype=hd5[path].attrs['dtype'])
            slices = (i, tm.channel_map[cm]) if dynamic else (tm.channel_map[cm],)
            tensor[slices] = np.count_nonzero(voltage == 0)
    return tensor


voltage_zeros = TensorMap(
    "voltage_zeros",
    interpretation=Interpretation.CONTINUOUS,
    path_prefix=PARTNERS_PREFIX,
    tensor_from_file=voltage_zeros,
    shape=(None, 12),
    channel_map=ECG_REST_AMP_LEADS,
    time_series_limit=0,
)


def v6_zeros_validator(tm: TensorMap, tensor: np.ndarray, hd5: h5py.File):
    voltage = decompress_data(data_compressed=hd5['V6'][()], dtype=hd5['V6'].attrs['dtype'])
    if np.count_nonzero(voltage == 0) > 10:
        raise ValueError(f'TensorMap {tm.name} has too many zeros in V6.')


partners_ecg_read_md_clean_supranodal_rhythms = TensorMap('partners_ecg_read_md_clean_supranodal_rhythms', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'retrograde_atrial_activation': 0, 'narrow_qrs_tachycardia': 1, 'unspecified': 2, 'svt': 3, 'ectopic_atrial_tachycardia': 4, 'ectopic_atrial_rhythm': 5, 'atrial_fibrillation': 6, 'atrial_flutter': 7, 'sinus_rhythm': 8}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'], 'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'], 'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'], 'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'], 'ectopic_atrial_tachycardia': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia'], 'narrow_qrs_tachycardia': ['narrow qrs tachycardia', 'narrow qrs tachycardia', 'tachycardia narrow qrs', 'narrow complex tachycardia'], 'retrograde_atrial_activation': ['retrograde atrial activation'], 'svt': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'], 'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm']}))

partners_ecg_read_pc_clean_supranodal_rhythms = TensorMap('partners_ecg_read_pc_clean_supranodal_rhythms', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'retrograde_atrial_activation': 0, 'narrow_qrs_tachycardia': 1, 'unspecified': 2, 'svt': 3, 'ectopic_atrial_tachycardia': 4, 'ectopic_atrial_rhythm': 5, 'atrial_fibrillation': 6, 'atrial_flutter': 7, 'sinus_rhythm': 8}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'], 'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'], 'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'], 'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'], 'ectopic_atrial_tachycardia': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia'], 'narrow_qrs_tachycardia': ['narrow qrs tachycardia', 'narrow qrs tachycardia', 'tachycardia narrow qrs', 'narrow complex tachycardia'], 'retrograde_atrial_activation': ['retrograde atrial activation'], 'svt': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'], 'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm']}))

partners_ecg_read_md_clean_sinus_rhythm = TensorMap('partners_ecg_read_md_clean_sinus_rhythm', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'sinus_arrhythmia': 0, 'unspecified': 1}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'sinus_arrhythmia': ['marked sinus arrhythmia', 'marked sinus arrhythmia', 'sinus arrhythmia', 'sinus arrhythmia']}))

partners_ecg_read_pc_clean_sinus_rhythm = TensorMap('partners_ecg_read_pc_clean_sinus_rhythm', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'sinus_arrhythmia': 0, 'unspecified': 1}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'sinus_arrhythmia': ['marked sinus arrhythmia', 'marked sinus arrhythmia', 'sinus arrhythmia', 'sinus arrhythmia']}))

partners_ecg_read_md_clean_atrial_flutter = TensorMap('partners_ecg_read_md_clean_atrial_flutter', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'fixed_block': 0, 'variable_block': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'unspecified': ['fibrillation/flutter', 'fibrillation/flutter', 'aflutter', 'aflutter', 'atrial flutter', 'atrial flutter', 'probable flutter', 'probable flutter', 'atrial flutter unspecified block', 'atrial flutter unspecified block'], 'fixed_block': ['atrial flutter fixed block', 'atrial flutter fixed block'], 'variable_block': ['atrial flutter variable block', 'atrial flutter variable block']}))

partners_ecg_read_pc_clean_atrial_flutter = TensorMap('partners_ecg_read_pc_clean_atrial_flutter', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'fixed_block': 0, 'variable_block': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'unspecified': ['fibrillation/flutter', 'fibrillation/flutter', 'aflutter', 'aflutter', 'atrial flutter', 'atrial flutter', 'probable flutter', 'probable flutter', 'atrial flutter unspecified block', 'atrial flutter unspecified block'], 'fixed_block': ['atrial flutter fixed block', 'atrial flutter fixed block'], 'variable_block': ['atrial flutter variable block', 'atrial flutter variable block']}))

partners_ecg_read_md_clean_ectopic_atrial_rhythm = TensorMap('partners_ecg_read_md_clean_ectopic_atrial_rhythm', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'wandering': 0, 'unifocal': 1, 'multifocal': 2, 'unspecified': 3}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'unifocal': ['unifocal atrial tachycardia', 'unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'unifocal ear', 'unusual p wave axis', 'unusual p wave axis', 'low atrial pacer', 'low atrial pacer'], 'multifocal': ['multifocal atrial tachycardia', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'multifocal ear', 'dual atrial foci ', 'dual atrial foci ', 'multiple atrial foci', 'multiple atrial foci', 'multifocal atrial rhythm', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'multifocal atrialrhythm'], 'wandering': ['wandering atrial tachycardia', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering ear', 'wandering atrial pacemaker', 'wandering atrial pacemaker'], 'unspecified': ['atrial rhythm', 'atrial rhythm', 'ectopic atrial rhythm', 'ectopic atrial rhythm', 'ear', 'ear', 'ectopic atrial bradycardia', 'ectopic atrial bradycardia', 'abnormal p vector', 'abnormal p vector', 'nonsinus atrial mechanism', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'low atrial bradycardia']}))

partners_ecg_read_pc_clean_ectopic_atrial_rhythm = TensorMap('partners_ecg_read_pc_clean_ectopic_atrial_rhythm', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'wandering': 0, 'unifocal': 1, 'multifocal': 2, 'unspecified': 3}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'unifocal': ['unifocal atrial tachycardia', 'unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'unifocal ear', 'unusual p wave axis', 'unusual p wave axis', 'low atrial pacer', 'low atrial pacer'], 'multifocal': ['multifocal atrial tachycardia', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'multifocal ear', 'dual atrial foci ', 'dual atrial foci ', 'multiple atrial foci', 'multiple atrial foci', 'multifocal atrial rhythm', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'multifocal atrialrhythm'], 'wandering': ['wandering atrial tachycardia', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering ear', 'wandering atrial pacemaker', 'wandering atrial pacemaker'], 'unspecified': ['atrial rhythm', 'atrial rhythm', 'ectopic atrial rhythm', 'ectopic atrial rhythm', 'ear', 'ear', 'ectopic atrial bradycardia', 'ectopic atrial bradycardia', 'abnormal p vector', 'abnormal p vector', 'nonsinus atrial mechanism', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'low atrial bradycardia']}))

partners_ecg_read_md_clean_ectopic_atrial_tachycardia = TensorMap('partners_ecg_read_md_clean_ectopic_atrial_tachycardia', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'unifocal': 0, 'multifocal': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'unspecified': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'unspecified ectopic atrial tachycardia'], 'unifocal': ['ectopic atrial tachycardia, unifocal', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'unifocal ectopic atrial tachycardia'], 'multifocal': ['ectopic atrial tachycardia, multifocal', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia', 'multifocal ectopic atrial tachycardia']}))

partners_ecg_read_pc_clean_ectopic_atrial_tachycardia = TensorMap('partners_ecg_read_pc_clean_ectopic_atrial_tachycardia', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'unifocal': 0, 'multifocal': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'unspecified': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'unspecified ectopic atrial tachycardia'], 'unifocal': ['ectopic atrial tachycardia, unifocal', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'unifocal ectopic atrial tachycardia'], 'multifocal': ['ectopic atrial tachycardia, multifocal', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia', 'multifocal ectopic atrial tachycardia']}))

partners_ecg_read_md_clean_svt = TensorMap('partners_ecg_read_md_clean_svt', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'avrt': 0, 'avnrt': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', dict_of_list = {'avnrt': ['av nodal reentry tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'avnrt'], 'avrt': ['atrioventricular reentrant tachycardia ', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt', 'avrt']}))

partners_ecg_read_pc_clean_svt = TensorMap('partners_ecg_read_pc_clean_svt', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'avrt': 0, 'avnrt': 1, 'unspecified': 2}, tensor_from_file=make_partners_ecg_label(keys='read_pc_clean', dict_of_list = {'avnrt': ['av nodal reentry tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'avnrt'], 'avrt': ['atrioventricular reentrant tachycardia ', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt', 'avrt']}))

partners_ecg_afib = TensorMap('partners_ecg_afib', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'no_atrial_fibrillation': 0, 'atrial_fibrillation': 1}, tensor_from_file=make_partners_ecg_label(keys='read_md_clean', not_found_key='no_atrial_fibrillation', dict_of_list = {'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation']}))
partners_ecg_afib_all = TensorMap('partners_ecg_afib_all', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'no_atrial_fibrillation': 0, 'atrial_fibrillation': 1}, tensor_from_file=make_partners_ecg_label(keys=['read_md_clean', 'read_pc_clean'], not_found_key='no_atrial_fibrillation', dict_of_list = {'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation']}))
partners_ecg_sinus_rhythm = TensorMap('partners_sinus_rhythm', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest', channel_map={'sinus_arrhythmia': 0, 'unspecified': 1}, tensor_from_file=make_partners_ecg_label(keys=['read_md_clean', 'read_pc_clean'], dict_of_list = {'sinus_arrhythmia': ['marked sinus arrhythmia', 'marked sinus arrhythmia', 'sinus arrhythmia', 'sinus arrhythmia']}))
partners_ecg_supranodal = TensorMap(
    'partners_ecg_supranodal', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest',
    channel_map={'retrograde_atrial_activation': 0, 'narrow_qrs_tachycardia': 1, 'unspecified': 2, 'supraventricular_tachycardia': 3, 'ectopic_atrial_tachycardia': 4, 'ectopic_atrial_rhythm': 5, 'atrial_fibrillation': 6, 'atrial_flutter': 7, 'sinus_rhythm': 8},
    tensor_from_file=make_partners_ecg_label(
        keys=['read_md_clean', 'read_pc_clean'], dict_of_list = {
            'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'],
            'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'],
            'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'],
            'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'],
            'ectopic_atrial_tachycardia': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia'],
            'narrow_qrs_tachycardia': ['narrow qrs tachycardia', 'narrow qrs tachycardia', 'tachycardia narrow qrs', 'narrow complex tachycardia'],
            'retrograde_atrial_activation': ['retrograde atrial activation'],
            'supraventricular_tachycardia': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'],
            'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm'],
        },
    ),
)

partners_ecg_supranodal_weighted = TensorMap(
    'partners_ecg_supranodal', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest',
    channel_map={'retrograde_atrial_activation': 0, 'narrow_qrs_tachycardia': 1, 'unspecified': 2, 'supraventricular_tachycardia': 3, 'ectopic_atrial_tachycardia': 4, 'ectopic_atrial_rhythm': 5, 'atrial_fibrillation': 6, 'atrial_flutter': 7, 'sinus_rhythm': 8},
    loss=weighted_crossentropy([20.0, 20.0, 10.0, 20.0, 20.0, 4.0, 1.0, 10.0, 1.0], 'partners_supranodal'),
    tensor_from_file=make_partners_ecg_label(
        keys=['read_md_clean', 'read_pc_clean'],
        dict_of_list = {
            'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'],
            'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'],
            'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'],
            'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'],
            'ectopic_atrial_tachycardia': ['ectopic atrial tachycardia', 'ectopic atrial tachycardia, unspecified', 'unspecified ectopic atrial tachycardia', 'ectopic atrial tachycardia, unifocal', 'unifocal ectopic atrial tachycardia', 'ectopic atrial tachycardia, multifocal', 'multifocal ectopic atrial tachycardia'],
            'narrow_qrs_tachycardia': ['narrow qrs tachycardia', 'narrow qrs tachycardia', 'tachycardia narrow qrs', 'narrow complex tachycardia'],
            'retrograde_atrial_activation': ['retrograde atrial activation'],
            'supraventricular_tachycardia': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'],
            'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm'],
        },
    ),
)

partners_ecg_supranodal_6 = TensorMap(
    'partners_ecg_supranodal_6', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest',
    channel_map={'supraventricular_tachycardia': 0, 'ectopic_atrial_rhythm': 1, 'atrial_flutter': 2, 'atrial_fibrillation': 3, 'sinus_rhythm': 4, 'unspecified': 5},
    tensor_from_file=make_partners_ecg_label(
        keys=['read_md_clean', 'read_pc_clean'],
        dict_of_list = {
            'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'],
            'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'],
            'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'],
            'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'],
            'supraventricular_tachycardia': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'], 'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm'],
        },
    ),
)

partners_ecg_supranodal_6_v6_valid = TensorMap(
    'partners_ecg_supranodal_6', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest',
    validator=v6_zeros_validator,
    channel_map={'supraventricular_tachycardia': 0, 'ectopic_atrial_rhythm': 1, 'atrial_flutter': 2, 'atrial_fibrillation': 3, 'sinus_rhythm': 4, 'unspecified': 5},
    tensor_from_file=make_partners_ecg_label(
        keys=['read_md_clean', 'read_pc_clean', 'testreason'],
        dict_of_list = {
            'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'],
            'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'],
            'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'],
            'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'],
            'supraventricular_tachycardia': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'], 'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm'],
        },
    ),
)

partners_ecg_supranodal_6_weighted = TensorMap(
    'partners_ecg_supranodal_6', interpretation=Interpretation.CATEGORICAL, time_series_limit=0, path_prefix='partners_ecg_rest',
    channel_map={'supraventricular_tachycardia': 0, 'ectopic_atrial_rhythm': 1, 'atrial_flutter': 2, 'atrial_fibrillation': 3, 'sinus_rhythm': 4, 'unspecified': 5},
    loss=weighted_crossentropy([70.0, 10.0, 50.0, 7.0, 1.0, 10.0], 'partners_supranodal_6'),
    tensor_from_file=make_partners_ecg_label(
        keys=['read_md_clean', 'read_pc_clean'],
        dict_of_list = {
            'atrial_fibrillation': ['atrial fib', 'afib', 'atrial fibrillation', 'atrial  fibrillation', 'afibrillation'],
            'sinus_rhythm': ['marked sinus arrhythmia', 'sinus arrhythmia', 'normal ecg', 'atrial bigeminal rhythm', 'atrial bigeminy and ventricular bigeminy', 'atrial trigeminy', 'atrialbigeminy', 'normal sinus rhythm', 'normal when compared with ecg of', 'rhythm has reverted to normal', 'rhythm is now clearly sinus', 'sinus bradycardia', 'sinus rhythm', 'sinus rhythm at a rate', 'sinus tachycardia', 'tracing within normal limits', 'tracing is within normal limits', 'tracing within normal limits', 'sinoatrial block', 'sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type i sinoatrial block', 'type i sa block', 'sinoatrial block, type ii', 'sa block, type i', 'type ii sinoatrial block', 'type ii sa block', 'sinus pause', 'sinus arrest', 'sinus slowing', 'with occasional native sinus beats', 'frequent native sinus beats', 'conducted sinus impulses', 'sinus mechanism has replaced', 'rhythm is normal sinus', 'rhythm remains normal sinus'],
            'atrial_flutter': ['fibrillation/flutter', 'aflutter', 'atrial flutter', 'probable flutter', 'atrial flutter fixed block', 'atrial flutter variable block', 'atrial flutter unspecified block', 'tachycardia possibly flutter'],
            'ectopic_atrial_rhythm': ['unifocal atrial tachycardia', 'unifocal ectopic atrial rhythm', 'unifocal ear', 'multifocal atrial tachycardia', 'multifocal ectopic atrial rhythm', 'multifocal ear', 'wandering atrial tachycardia', 'wandering ectopic atrial rhythm', 'wandering ear', 'wandering atrial pacemaker', 'atrial rhythm', 'ectopic atrial rhythm', 'ear', 'dual atrial foci ', 'ectopic atrial bradycardia', 'multiple atrial foci', 'abnormal p vector', 'nonsinus atrial mechanism', 'p wave axis suggests atrial rather than sinus mechanism', 'low atrial bradycardia', 'multifocal atrial rhythm', 'multifocal atrialrhythm', 'unusual p wave axis', 'low atrial pacer'],
            'supraventricular_tachycardia': ['sinus arrhythmia accelerated atrioventricular junctional rhythm', 'supraventricular tachycardia', 'accelerated atrioventricular nodal rhythm', 'accelerated nodal rhythm', 'atrial tachycardia', 'av nodal reentry tachycardia', 'atrioventricular nodal reentry tachycardia', 'avnrt', 'atrioventricular reentrant tachycardia ', 'av reentrant tachycardia ', 'avrt'], 'unspecified': ['junctional tachycardia', 'atrial arrhythmia', 'technically poor tracing ', 'accelerated idioventricular rhythm', 'atrial activity is indistinct', 'rhythm uncertain', 'rhythm unclear', 'uncertain rhythm', 'undetermined rhythm', 'supraventricular rhythm'],
        },
    ),
)

patientage = TensorMap(
    'patientage', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'patientage': 0},
)

af_in_read = TensorMap(
    'af_in_read', Interpretation.CATEGORICAL, path_prefix='categorical',
    channel_map={'no_atrial_fibrillation': 0, 'atrial_fibrillation': 1},
)

is_female = TensorMap(
    'is_female', Interpretation.CATEGORICAL, path_prefix='categorical', annotation_units=2,
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy',
)



import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()
# Get the requirements from the requirements file
requirements = (here / 'docker/vm_boot_images/config/tensorflow-requirements.txt').read_text(encoding='utf-8')
long_description = (here / 'README.md').read_text(encoding='utf-8')


setup(
    name='ml4h',
    version='0.0.19',
    description='Machine Learning for Health python package',
    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',
    url='https://github.com/broadinstitute/ml4h',
    python_requires='>=3.6',
    install_requires=requirements,
    packages=find_packages(),
)



import os
import csv
import logging
import h5py
import numpy as np
import pandas as pd
from typing import List, Tuple, Dict
from ml4h.TensorMap import TensorMap, Interpretation
from ml4h.normalizer import Standardize


def tensor_path(path_prefix: str, name: str) -> str:
    """
    In the future, TMAPs should be generated using this same function
    """
    return f'/{path_prefix}/{name}/'


def tensor_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
    return np.array(hd5[tm.name])

def named_tensor_from_hd5(name):
    def _tensor_from_hd5(tm: TensorMap, hd5: h5py.File, dependents: Dict = {}) -> np.ndarray:
        return np.array(hd5[name])
    return _tensor_from_hd5


def all_dates(hd5: h5py.File, path_prefix: str, name: str) -> List[str]:
    """
    Gets the dates in the hd5 with path_prefix, dtype, name.
    """
    # TODO: This ideally would be implemented to not depend on the order of name,
    # date, dtype, path_prefix in the hd5s. Unfortunately, that's hard to do
    # efficiently
    return hd5[path_prefix][name]


def pass_nan(tensor):
    return tensor


def fail_nan(tensor):
    if np.isnan(tensor).any():
        raise ValueError('Tensor contains nans.')
    return tensor


def nan_to_mean(tensor, max_allowed_nan_fraction=.2):
    tensor_isnan = np.isnan(tensor)
    if np.count_nonzero(tensor_isnan) / tensor.size > max_allowed_nan_fraction:
        raise ValueError('Tensor contains too many nans.')
    tensor[tensor_isnan] = np.nanmean(tensor)
    return tensor


def get_tensor_at_first_date(
    hd5: h5py.File,
    path_prefix: str,
    name: str,
    handle_nan=fail_nan,
):
    """
    Gets the numpy array at the first date of path_prefix, dtype, name.
    """
    dates = all_dates(hd5, path_prefix, name)
    if not dates:
        raise ValueError(f'No {name} values values available.')
    tensor = np.array(
        hd5[f'{tensor_path(path_prefix=path_prefix, name=name)}{min(dates)}/'],
        dtype=np.float32,
    )
    tensor = handle_nan(tensor)
    return tensor

def get_tensor_at_last_date(
    hd5: h5py.File,
    path_prefix: str,
    name: str,
    handle_nan=fail_nan,
):
    """
    Gets the numpy array at the last date of path_prefix, dtype, name.
    """
    dates = all_dates(hd5, path_prefix, name)
    if not dates:
        raise ValueError(f'No {name} values values available.')
    tensor = np.array(
        hd5[f'{tensor_path(path_prefix=path_prefix, name=name)}{max(dates)}/'],
        dtype=np.float32,
    )
    tensor = handle_nan(tensor)
    return tensor


def pad_or_crop_array_to_shape(new_shape: Tuple, original: np.ndarray):
    if new_shape == original.shape:
        return original
    result = np.zeros(new_shape)
    slices = tuple(
        slice(min(original.shape[i], new_shape[i]))
        for i in range(len(original.shape))
    )

    # Allow expanding one dimension eg (256, 256) can become (256, 256, 1)
    if len(new_shape) - len(original.shape) == 1:
        padded = result[..., 0]
    else:
        padded = result

    padded[slices] = original[slices]
    return result


def normalized_first_date(tm: TensorMap, hd5: h5py.File, dependents=None):
    tensor = get_tensor_at_first_date(hd5, tm.path_prefix, tm.name)
    if tm.axes() > 1:
        return pad_or_crop_array_to_shape(tm.shape, tensor)
    else:
        return tensor


def build_tensor_from_file(
    file_name: str,
    target_column: str,
    normalization: bool = False,
):
    """
    Build a tensor_from_file function from a column in a file.
    Only works for continuous values.
    When normalization is True values will be normalized according to the mean and std of all of the values in the column.
    """
    error = None
    try:
        with open(file_name, 'r') as f:
            ext = file_name.split('.')[1]
            delimiter = ',' if ext == 'csv' else '\t'
            reader = csv.reader(f, delimiter=delimiter)
            header = next(reader)
            index = header.index(target_column)
            table = {}
            for row in reader:
                try:
                    table[row[0]] = np.array([float(row[index])])
                except ValueError:
                    logging.debug(f'ValueError parsing: {row[index]}')
            if normalization:
                value_array = np.array(
                    [sub_array[0] for sub_array in table.values()],
                )
                mean = value_array.mean()
                std = value_array.std()
                logging.info(
                    f'Normalizing TensorMap from file {file_name}, column {target_column} with mean: '
                    f'{mean:.2f}, std: {std:.2f}', )
    except FileNotFoundError as e:
        error = e

    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if error:
            raise error
        if normalization:
            tm.normalization = Standardize(mean=mean, std=std)
        try:
            return table[
                os.path.basename(hd5.filename).replace('.hd5', '')
            ].copy()
        except KeyError:
            raise KeyError(f'Sample id not in file {file_name}.')

    return tensor_from_file


def build_categorical_tensor_from_file(
    file_name: str,
    target_column: str,
):
    """
    Build a tensor_from_file function from a column in a file as categorical classifier.
    """
    error = None
    try:
        ext = file_name.split('.')[1]
        delimiter = ',' if ext == 'csv' else '\t'
        df = pd.read_csv(file_name, delimiter=delimiter)
        table = dict(zip(df[df.columns[0]].tolist(), df[target_column].tolist()))
        logging.info(f'Categorical table from column {target_column} counts:\n{df[target_column].value_counts()}')
    except FileNotFoundError as e:
        error = e

    def tensor_from_file(tm: TensorMap, hd5: h5py.File, dependents=None):
        if error:
            raise error
        try:
            tensor = np.zeros(tm.shape, dtype=np.float32)
            val = table[int(os.path.basename(hd5.filename).replace('.hd5', ''))]
            tensor[tm.channel_map[val]] = 1.0
            return tensor
        except KeyError as e:
            raise KeyError(f'Sample id not in file {file_name}, Error: {e}.')

    return tensor_from_file



import h5py
import numpy as np
from datetime import datetime
import logging
from typing import List, Tuple

from ml4h.normalizer import Standardize, ZeroMeanStd1
from ml4h.tensormap.general import tensor_path
from ml4h.TensorMap import TensorMap, Interpretation, str2date, make_range_validator
from ml4h.defines import StorageType


def is_genetic_man(hd5):
    return 'Genetic-sex_Male_0_0' in hd5['categorical']


def is_genetic_woman(hd5):
    return 'Genetic-sex_Female_0_0' in hd5['categorical']


def age_in_years_tensor(
    date_key,
    birth_key='continuous/34_Year-of-birth_0_0',
    population_normalize=False,
):
    def age_at_tensor_from_file(
        tm: TensorMap,
        hd5: h5py.File,
        dependents=None,
    ):
        try:
            age = np.array([hd5['ecg/latest/patient_info/Age'][()]])
        except:
            logging.info('could not get age')
            raise KeyError('cold not')
        # age = age.astype("float")

        return age
        # return tm.normalize_and_validate(np.array([assess_date.year-birth_year]))

    return age_at_tensor_from_file


def prevalent_incident_tensor(start_date_key, event_date_key):
    def _prevalent_incident_tensor_from_file(
        tm: TensorMap,
        hd5: h5py.File,
        dependents=None,
    ):
        index = 0
        categorical_data = np.zeros(tm.shape, dtype=np.float32)
        if tm.hd5_key_guess() in hd5:
            data = tm.hd5_first_dataset_in_group(hd5, tm.hd5_key_guess())
            if tm.storage_type == StorageType.CATEGORICAL_INDEX or tm.storage_type == StorageType.CATEGORICAL_FLAG:
                index = int(data[0])
                categorical_data[index] = 1.0
            else:
                categorical_data = np.array(data)
        elif tm.storage_type == StorageType.CATEGORICAL_FLAG:
            categorical_data[index] = 1.0
        else:
            raise ValueError(
                f"No HD5 Key at prefix {tm.path_prefix} found for tensor map: {tm.name}.",
            )

        if index != 0:
            if event_date_key in hd5 and start_date_key in hd5:
                disease_date = str2date(str(hd5[event_date_key][0]))
                assess_date = str2date(str(hd5[start_date_key][0]))
            else:
                raise ValueError(f"No date found for tensor map: {tm.name}.")
            index = 1 if disease_date < assess_date else 2
        categorical_data[index] = 1.0
        return categorical_data

    return _prevalent_incident_tensor_from_file


def prevalent_tensor(start_date_key: str, event_date_key: str, start_date_is_attribute: bool = False):
    def _prevalent_tensor_from_file(
        tm: TensorMap,
        hd5: h5py.File,
        dependents=None,
    ):
        index = 0
        categorical_data = np.zeros(tm.shape, dtype=np.float32)
        if tm.hd5_key_guess() in hd5:
            data = tm.hd5_first_dataset_in_group(hd5, tm.hd5_key_guess())
            if tm.storage_type == StorageType.CATEGORICAL_INDEX or tm.storage_type == StorageType.CATEGORICAL_FLAG:
                index = int(data[0])
                categorical_data[index] = 1.0
            else:
                categorical_data = np.array(data)
        elif tm.storage_type == StorageType.CATEGORICAL_FLAG:
            categorical_data[index] = 1.0
        else:
            raise ValueError(
                f"No HD5 Key at prefix {tm.path_prefix} found for tensor map: {tm.name}.",
            )

        if index != 0:
            if event_date_key in hd5 and start_date_key in hd5:
                disease_date = str2date(str(hd5[event_date_key][0]))
                if start_date_is_attribute:
                    assess_date = datetime.utcfromtimestamp(hd5[start_date_key].attrs['date']).date()
                else:
                    assess_date = str2date(str(hd5[start_date_key][0]))
            else:
                raise ValueError(f"No date found for tensor map: {tm.name}.")
            index = 1 if disease_date < assess_date else 0
        categorical_data[index] = 1.0
        return categorical_data
    return _prevalent_tensor_from_file


def preprocess_with_function(fxn, hd5_key=None):
    def preprocess_tensor_from_file(tm, hd5, dependents={}):
        missing = True
        continuous_data = np.zeros(tm.shape, dtype=np.float32)
        my_key = tm.hd5_key_guess() if hd5_key is None else hd5_key
        if my_key in hd5:
            missing = False
            continuous_data[0] = tm.hd5_first_dataset_in_group(hd5, my_key)[0]
        if missing and tm.sentinel is None:
            raise ValueError(
                f'No value found for {tm.name}, a continuous TensorMap with no sentinel value, and channel keys:{list(tm.channel_map.keys())}.',
            )
        elif missing:
            continuous_data[:] = tm.sentinel
        return fxn(continuous_data)

    return preprocess_tensor_from_file


def _weekly_alcohol(instance):
    alcohol_keys = [
        f'1568_Average-weekly-red-wine-intake_{instance}_0',
        f'1578_Average-weekly-champagne-plus-white-wine-intake_{instance}_0',
        f'1588_Average-weekly-beer-plus-cider-intake_{instance}_0',
        f'1598_Average-weekly-spirits-intake_{instance}_0',
        f'1608_Average-weekly-fortified-wine-intake_{instance}_0',
    ]

    def alcohol_from_file(tm, hd5, dependents={}):
        drinks = 0
        for k in alcohol_keys:
            data = tm.hd5_first_dataset_in_group(
                hd5, key_prefix=f'{tm.path_prefix}/{k}',
            )
            drinks += float(data[0])
        return np.array([drinks], dtype=np.float32)

    return alcohol_from_file


log_25781_2 = TensorMap(
    '25781_Total-volume-of-white-matter-hyperintensities-from-T1-and-T2FLAIR-images_2_0',
    loss='log_cosh',
    path_prefix='continuous',
    normalization={
        'mean': 7,
        'std': 8,
    },
    tensor_from_file=preprocess_with_function(np.log),
    channel_map={'white-matter-hyper-intensities': 0},
)

weight_lbs_2 = TensorMap(
    'weight_lbs',
    Interpretation.CONTINUOUS,
    normalization={
        'mean': 168.74,
        'std': 34.1,
    },
    loss='log_cosh',
    channel_map={'weight_lbs': 0},
    tensor_from_file=preprocess_with_function(
        lambda x: x * 2.20462,
        'continuous/21002_Weight_2_0',
    ),
)

weekly_alcohol_0 = TensorMap(
    'weekly_alcohol_0',
    loss='log_cosh',
    path_prefix='continuous',
    channel_map={'weekly_alcohol_0': 0},
    tensor_from_file=_weekly_alcohol(0),
)
weekly_alcohol_1 = TensorMap(
    'weekly_alcohol_1',
    loss='log_cosh',
    path_prefix='continuous',
    channel_map={'weekly_alcohol_1': 0},
    tensor_from_file=_weekly_alcohol(1),
)
weekly_alcohol_2 = TensorMap(
    'weekly_alcohol_2',
    loss='log_cosh',
    path_prefix='continuous',
    channel_map={'weekly_alcohol_2': 0},
    tensor_from_file=_weekly_alcohol(2),
)

###
weight_kg = TensorMap('weight_kg',  Interpretation.CONTINUOUS, normalization={'mean': 76.54286701805927, 'std': 15.467605416933122}, loss='log_cosh', channel_map={'weight_kg': 0})
height_cm = TensorMap('height_cm',  Interpretation.CONTINUOUS, normalization={'mean': 169.18064748408653, 'std': 9.265265197273026}, loss='log_cosh', channel_map={'height_cm': 0})
bmi_bsa = TensorMap('bmi',  Interpretation.CONTINUOUS, normalization={'mean': 26.65499238706321, 'std': 4.512077188749083}, loss='log_cosh', channel_map={'bmi': 0})

mothers_age = TensorMap(
    'mothers_age_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'mother_age': 0, 'mother_alive': 2, 'mother_dead': 3, 'not-missing': 1},
    normalization={'mean': 75.555, 'std': 11.977}, annotation_units = 4,
)

fathers_age = TensorMap(
    'fathers_age_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'father_age': 0, 'father_alive': 2, 'father_dead': 3, 'not-missing': 1},
    normalization={'mean':70.928, 'std': 12.746}, annotation_units = 4,
)

genetic_sex = TensorMap(
    'Genetic-sex_Male_0_0', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    path_prefix='categorical', annotation_units=2, loss='categorical_crossentropy',
    channel_map={'Genetic-sex_Female_0_0': 0, 'Genetic-sex_Male_0_0': 1},
)
partition_i = 128
a_units = 128
genetic_sex_partition = TensorMap(
    'Genetic-sex_Male_0_0', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    path_prefix='categorical', days_window=partition_i, annotation_units=a_units, loss='categorical_crossentropy',
    channel_map={'Genetic-sex_Female_0_0': 0, 'Genetic-sex_Male_0_0': 1},
)
age_2_partition = TensorMap(
    '21003_Age-when-attended-assessment-centre_2_0', Interpretation.CONTINUOUS, days_window=partition_i,
    annotation_units=a_units, path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.358, std=7.555),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)
sex = TensorMap(
    'Sex_Male_0_0', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy', annotation_units=2,
)

sex_mgb = TensorMap(
    'sex', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy', annotation_units=2,
)

is_male_mgb = TensorMap(
    'is_male', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy',
)

age_in_days = TensorMap(
    'age_in_days', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh',
    #normalization=Standardize(mean=65, std=(1/365.0)),
    normalization=ZeroMeanStd1(),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)
bmi = TensorMap(
    '23104_Body-mass-index-BMI_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'23104_Body-mass-index-BMI_0_0': 0}, validator=make_range_validator(0, 100),
    normalization={'mean': 27.432, 'std': 4.785},
)
bmi_ukb = TensorMap(
    'bmi', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'23104_Body-mass-index-BMI_0_0': 0}, annotation_units=1,
    validator=make_range_validator(0, 300), normalization={'mean': 27.432061533712652, 'std': 4.785244772462738}, loss='log_cosh',
)
bmi_2 = TensorMap(
    '21001_Body-mass-index-BMI_2_0', Interpretation.CONTINUOUS, path_prefix='continuous',  loss='log_cosh',
    channel_map={'21001_Body-mass-index-BMI_2_0': 0}, validator=make_range_validator(0, 300),
    normalization=Standardize(mean=27.3397, std=4.7721),
)
bmi_2_partition = TensorMap(
    '21001_Body-mass-index-BMI_2_0', Interpretation.CONTINUOUS, path_prefix='continuous',  loss='log_cosh',
    channel_map={'21001_Body-mass-index-BMI_2_0': 0}, validator=make_range_validator(0, 300),
    normalization=Standardize(mean=27.3397, std=4.7721), days_window=partition_i, annotation_units=a_units,
)
bmi_21_0 = TensorMap(
    '21001_Body-mass-index-BMI_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'21001_Body-mass-index-BMI_0_0': 0}, validator=make_range_validator(0, 300),
    normalization=Standardize(mean=27.3397, std=4.7721),
)
birth_year = TensorMap(
    '22200_Year-of-birth_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'22200_Year-of-birth_0_0': 0}, annotation_units=1, loss='log_cosh',
    validator=make_range_validator(1901, 2025), normalization={'mean': 1952.0639129359386, 'std': 7.656326148519739},
)
birth_year_34 = TensorMap(
    '34_Year-of-birth_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'34_Year-of-birth_0_0': 0}, annotation_units=1, loss='log_cosh',
    validator=make_range_validator(1901, 2025), normalization = {'mean': 1952.0639129359386, 'std': 7.656326148519739},
)
age_0 = TensorMap(
    '21003_Age-when-attended-assessment-centre_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization={'mean': 56.52847159208494, 'std': 8.095287610193827}, channel_map={'21003_Age-when-attended-assessment-centre_0_0': 0},
)
age_1 = TensorMap(
    '21003_Age-when-attended-assessment-centre_1_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization={'mean': 61.4476555588322, 'std': 7.3992113757847005}, channel_map={'21003_Age-when-attended-assessment-centre_1_0': 0},
)
age_2 = TensorMap(
    '21003_Age-when-attended-assessment-centre_2_0', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)

age_2_wide = TensorMap(
    'age_from_wide_csv', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)
age_2_patientage = TensorMap(
    'patientage', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)

age_2_wide = TensorMap(
    'age_from_wide_csv', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)
age_2_patientage = TensorMap(
    'patientage', Interpretation.CONTINUOUS,
    path_prefix='continuous', loss='log_cosh', validator=make_range_validator(1, 120),
    normalization=Standardize(mean=63.35798891483556, std=7.554638350423902),
    channel_map={'21003_Age-when-attended-assessment-centre_2_0': 0},
)

af_dummy = TensorMap(
    'af_in_read', Interpretation.CATEGORICAL, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG,
    channel_map={'no_atrial_fibrillation': 0, 'atrial_fibrillation': 1},
)
af_dummy2 = TensorMap(
    'af_in_read', Interpretation.CATEGORICAL, path_prefix='categorical', storage_type=StorageType.CATEGORICAL_FLAG,
    channel_map={'no_af_in_read': 0, 'af_in_read': 1},
)
sex_dummy = TensorMap(
    'sex_from_wide', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    path_prefix='categorical', annotation_units=2,
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy',
)
sex_dummy1 = TensorMap(
    'sex', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    path_prefix='categorical', annotation_units=2,
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy',
)
sex_dummy2 = TensorMap(
    'is_female', Interpretation.CATEGORICAL,
     path_prefix='categorical', annotation_units=2,
    channel_map={'Sex_Female_0_0': 0, 'Sex_Male_0_0': 1}, loss='categorical_crossentropy',
)
sex_dummy3 = TensorMap(
    'sex_from_wide', Interpretation.CATEGORICAL, storage_type=StorageType.CATEGORICAL_FLAG,
    path_prefix='categorical', annotation_units=2,
    channel_map={'female': 0, 'male': 1}, loss='categorical_crossentropy',
)
brain_volume = TensorMap(
    '25010_Volume-of-brain-greywhite-matter_2_0', Interpretation.CONTINUOUS, path_prefix='continuous', normalization={'mean': 1165940.0, 'std': 111511.0},
    channel_map={'25010_Volume-of-brain-greywhite-matter_2_0': 0}, loss='log_cosh', loss_weight=0.1,
)

sodium = TensorMap(
    '30530_Sodium-in-urine_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'30530_Sodium-in-urine_0_0': 0},
    normalization={'mean': 77.45323967267045, 'std': 44.441236848463774}, annotation_units=1, loss='log_cosh',
)
potassium = TensorMap(
    '30520_Potassium-in-urine_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'30520_Potassium-in-urine_0_0': 0},
    normalization={'mean': 63.06182700345117, 'std': 33.84208704773539}, annotation_units=1, loss='log_cosh',
)
cholesterol_hdl = TensorMap(
    '30760_HDL-cholesterol_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'30760_HDL-cholesterol_0_0': 0},
    normalization={'mean': 1.4480129055069355, 'std': 0.3823115953478376}, annotation_units=1, loss='log_cosh',
)
cholesterol = TensorMap(
    '30690_Cholesterol_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'30690_Cholesterol_0_0': 0},
    normalization={'mean': 5.692381214399044, 'std': 1.1449409331668705}, annotation_units=1, loss='log_cosh',
)

cigarettes = TensorMap('2887_Number-of-cigarettes-previously-smoked-daily_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'2887_Number-of-cigarettes-previously-smoked-daily_0_0': 0}, normalization = {'mean': 18.92662147068755, 'std':10.590930376362259}, annotation_units=1)
alcohol = TensorMap('5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'5364_Average-weekly-intake-of-other-alcoholic-drinks_0_0': 0}, normalization = {'mean': 0.03852570253005904, 'std':0.512608370266108}, annotation_units=1)


def alcohol_channel_map(instance=0, array_idx=0):
    return {
        f'Alcohol-intake-frequency_Never_{instance}_{array_idx}': 0,
        f'Alcohol-intake-frequency_Special-occasions-only_{instance}_{array_idx}': 1,
        f'Alcohol-intake-frequency_One-to-three-times-a-month_{instance}_{array_idx}': 2,
        f'Alcohol-intake-frequency_Once-or-twice-a-week_{instance}_{array_idx}': 3,
        f'Alcohol-intake-frequency_Three-or-four-times-a-week_{instance}_{array_idx}': 4,
        f'Alcohol-intake-frequency_Daily-or-almost-daily_{instance}_{array_idx}': 5,
    }


alcohol_0 = TensorMap('alcohol_0', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_channel_map(instance=0))
alcohol_1 = TensorMap('alcohol_1', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_channel_map(instance=1))
alcohol_2 = TensorMap('alcohol_2', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_channel_map(instance=2))


def alcohol_status_map(instance=0, array_idx=0):
    return {
        f'Alcohol-drinker-status_Never_{instance}_{array_idx}': 0,
        f'Alcohol-drinker-status_Previous_{instance}_{array_idx}': 1,
        f'Alcohol-drinker-status_Current_{instance}_{array_idx}': 2,
    }


alcohol_status_0 = TensorMap('alcohol_status_0', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_status_map(instance=0))
alcohol_status_1 = TensorMap('alcohol_status_1', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_status_map(instance=1))
alcohol_status_2 = TensorMap('alcohol_status_2', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_status_map(instance=2))


def alcohol_meals_map(instance=0, array_idx=0):
    return {
        f'Alcohol-usually-taken-with-meals_No_{instance}_{array_idx}': 0,
        f'Alcohol-usually-taken-with-meals_It-varies_{instance}_{array_idx}': 1,
        f'Alcohol-usually-taken-with-meals_Yes_{instance}_{array_idx}': 2,
    }


alcohol_meals_0 = TensorMap('alcohol_meals_0', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_meals_map(instance=0))
alcohol_meals_1 = TensorMap('alcohol_meals_1', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_meals_map(instance=1))
alcohol_meals_2 = TensorMap('alcohol_meals_2', Interpretation.CATEGORICAL, path_prefix='categorical', channel_map=alcohol_meals_map(instance=2))

coffee = TensorMap(
    '1498_Coffee-intake_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'1498_Coffee-intake_0_0': 0},
    normalization={'mean': 2.015086529948216, 'std': 2.0914960998390497}, annotation_units=1,
)
water = TensorMap(
    '1528_Water-intake_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'1528_Water-intake_0_0': 0},
    normalization={'mean': 2.7322977785723324, 'std': 2.261996814128837}, annotation_units=1,
)
meat = TensorMap(
    '3680_Age-when-last-ate-meat_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'3680_Age-when-last-ate-meat_0_0': 0},
    normalization={'mean': 29.74062983480561, 'std': 14.417292213873964}, annotation_units=1,
)
walks = TensorMap(
    '864_Number-of-daysweek-walked-10-minutes_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'864_Number-of-daysweek-walked-10-minutes_0_0': 0},
    normalization={'mean': 5.369732285440756, 'std': 1.9564911925721618}, annotation_units=1,
)
walk_duration = TensorMap(
    '874_Duration-of-walks_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'874_Duration-of-walks_0_0': 0},
    normalization={'mean': 61.64092215093373, 'std': 78.79522990818906}, annotation_units=1,
)
physical_activities = TensorMap(
    '884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'884_Number-of-daysweek-of-moderate-physical-activity-10-minutes_0_0': 0},
    normalization={'mean': 3.6258833281089258, 'std': 2.3343738999823676}, annotation_units=1,
)
physical_activity = TensorMap(
    '894_Duration-of-moderate-activity_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'894_Duration-of-moderate-activity_0_0': 0},
    normalization={'mean': 66.2862593866103, 'std': 77.28681218835422}, annotation_units=1,
)
physical_activity_vigorous = TensorMap(
    '904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0', Interpretation.CONTINUOUS,
    channel_map={'904_Number-of-daysweek-of-vigorous-physical-activity-10-minutes_0_0': 0}, path_prefix='continuous',
    normalization={'mean': 1.838718301735063, 'std': 1.9593505421480895}, annotation_units=1,
)
physical_activity_vigorous_duration = TensorMap(
    '914_Duration-of-vigorous-activity_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'914_Duration-of-vigorous-activity_0_0': 0},
    normalization={'mean': 44.854488382965144, 'std': 48.159967071781466}, annotation_units=1,
)
tv = TensorMap(
    '1070_Time-spent-watching-television-TV_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'1070_Time-spent-watching-television-TV_0_0': 0},
    normalization={'mean': 2.7753595642790914, 'std': 1.7135478462887321}, annotation_units=1,
)
computer = TensorMap(
    '1080_Time-spent-using-computer_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'1080_Time-spent-using-computer_0_0': 0},
    normalization={'mean': 0.9781465855433753, 'std': 1.4444414103121512}, annotation_units=1,
)
car = TensorMap(
    '1090_Time-spent-driving_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', channel_map={'1090_Time-spent-driving_0_0': 0},
    normalization={'mean': 0.8219851505445748, 'std': 1.304094814200189}, annotation_units=1,
)
summer = TensorMap(
    '1050_Time-spend-outdoors-in-summer_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'1050_Time-spend-outdoors-in-summer_0_0': 0},
    normalization={'mean': 3.774492304870845, 'std': 2.430483731404539}, annotation_units=1,
)
winter = TensorMap(
    '1060_Time-spent-outdoors-in-winter_0_0', Interpretation.CONTINUOUS, path_prefix='continuous',
    channel_map={'1060_Time-spent-outdoors-in-winter_0_0': 0},
    normalization={'mean': 1.8629686916635555, 'std': 1.88916218603397}, annotation_units=1,
)

systolic_blood_pressure_0 = TensorMap(
    '4080_Systolic-blood-pressure-automated-reading_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4080_Systolic-blood-pressure-automated-reading_0_0': 0}, validator=make_range_validator(40, 400),
    normalization={'mean': 137.79964191990328, 'std': 19.292863700283757},
)
diastolic_blood_pressure_0 = TensorMap(
    '4079_Diastolic-blood-pressure-automated-reading_0_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4079_Diastolic-blood-pressure-automated-reading_0_0': 0}, validator=make_range_validator(20, 300),
    normalization={'mean': 82.20657551284782, 'std': 10.496040770224475},
)

systolic_blood_pressure_1 = TensorMap(
    '4080_Systolic-blood-pressure-automated-reading_1_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4080_Systolic-blood-pressure-automated-reading_1_0': 0}, validator=make_range_validator(40, 400),
    normalization={'mean': 137.79964191990328, 'std': 19.292863700283757},
)
diastolic_blood_pressure_1 = TensorMap(
    '4079_Diastolic-blood-pressure-automated-reading_1_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4079_Diastolic-blood-pressure-automated-reading_1_0': 0}, validator=make_range_validator(20, 300),
    normalization={'mean': 82.20657551284782, 'std': 10.496040770224475},
)

systolic_blood_pressure_2 = TensorMap(
    '4080_Systolic-blood-pressure-automated-reading_2_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4080_Systolic-blood-pressure-automated-reading_2_0': 0}, validator=make_range_validator(40, 400),
    normalization={'mean': 137.79964191990328, 'std': 19.292863700283757},
)
diastolic_blood_pressure_2 = TensorMap(
    '4079_Diastolic-blood-pressure-automated-reading_2_0', Interpretation.CONTINUOUS, path_prefix='continuous', loss='log_cosh',
    channel_map={'4079_Diastolic-blood-pressure-automated-reading_2_0': 0}, validator=make_range_validator(20, 300),
    normalization={'mean': 82.20657551284782, 'std': 10.496040770224475},
)

hypertension = TensorMap(
    'hypertension', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_hypertension': 0, 'hypertension': 1},
)

hypertension_diagnosis = TensorMap(
    'hypertension_diagnosis', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_hypertension': 0, 'hypertension': 1},
)

hypertension_icd_bp = TensorMap(
    'hypertension_icd_bp', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_INDEX, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_hypertension': 0, 'hypertension': 1},
)
htn_icd_bp = TensorMap(
    'htn_icd_bp', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_INDEX, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_hypertension': 0, 'hypertension': 1},
)

diabetes = TensorMap(
    'dm', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_INDEX, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_diabetes_type_2': 0, 'diabetes_type_2': 1},
)

hypercholesterolemia = TensorMap(
    'hypercholesterolemia', Interpretation.CATEGORICAL,
    storage_type=StorageType.CATEGORICAL_INDEX, path_prefix='categorical',
    loss='categorical_crossentropy',
    channel_map={'no_hypercholesterolemia': 0, 'hypercholesterolemia': 1},
)

hypertension_med = TensorMap(
    'start_fu_hypertension_med', Interpretation.CATEGORICAL, loss='categorical_crossentropy',
    storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    channel_map={'no_start_fu_hypertension_med':0, 'start_fu_hypertension_med': 1},
)
peak_vo2 = TensorMap(
    'peak_vo2', Interpretation.CONTINUOUS,
    storage_type=StorageType.CATEGORICAL_INDEX, path_prefix='continuous',
    channel_map={'peak_vo2':0},
)
cad = TensorMap(
    'cad', Interpretation.CATEGORICAL, loss='categorical_crossentropy',
    storage_type=StorageType.CATEGORICAL_FLAG, path_prefix='categorical',
    channel_map={'no_coronary_artery_disease':0, 'coronary_artery_disease': 1},
)

categorical_phenotypes_25 = TensorMap(
    'categorical-phenotypes-25', Interpretation.CATEGORICAL, path_prefix='categorical',
    channel_map={
        'Adopted-as-a-child_No_0_0': 0,
        'Beef-intake_Less-than-once-a-week_0_0': 1,
        'Breastfed-as-a-baby_Yes_0_0': 2,
        'Country-of-birth-UKelsewhere_England_0_0': 3,
        'Current-tobacco-smoking_No_0_0': 4,
        'Drive-faster-than-motorway-speed-limit_Neverrarely_0_0': 5,
        'Fracturedbroken-bones-in-last-5-years_No_0_0': 6,
        'Genetic-sex_Female_0_0': 7,
        'Hearing-difficultyproblems_No_0_0': 8,
        'Major-dietary-changes-in-the-last-5-years_No_0_0': 9,
        'Processed-meat-intake_Less-than-once-a-week_0_0': 10,
        'Processed-meat-intake_Once-a-week_0_0': 11,
        'Salt-added-to-food_Neverrarely_0_0': 12,
        'Salt-added-to-food_Sometimes_0_0': 13,
        'Shortness-of-breath-walking-on-level-ground_No_0_0': 14,
        'Smoking-status_Never_0_0': 15,
        'Smokingsmokers-in-household_No_0_0': 16,
        'Smoking-status_Previous_0_0': 17,
        'Usual-walking-pace_Brisk-pace_0_0': 18,
        'Usual-walking-pace_Steady-average-pace_0_0': 19,
        'Variation-in-diet_Sometimes_0_0': 20,
        'Vitamin-and-mineral-supplements_None-of-the-above_0_0': 21,
        'Wears-glasses-or-contact-lenses_Yes_0_0': 22,
        'Weight-change-compared-with-1-year-ago_No-weigh-about-the-same_0_0': 23,
        'Weight-change-compared-with-1-year-ago_Yes-gained-weight_0_0': 24,

    },
)

categorical_phenotypes_36 = TensorMap(
    'categorical-phenotypes-36', Interpretation.CATEGORICAL, path_prefix='categorical',
    channel_map={
        'Adopted-as-a-child_No_0_0': 0,
        'Breastfed-as-a-baby_Yes_0_0': 1,
        'Country-of-birth-UKelsewhere_England_0_0': 2,
        'Current-tobacco-smoking_No_0_0': 3,
        'Drive-faster-than-motorway-speed-limit_Neverrarely_0_0': 4,
        'Facial-ageing_Younger-than-you-are_0_0': 5,
        'Father-still-alive_No_0_0': 6,
        'Fracturedbroken-bones-in-last-5-years_No_0_0': 7,
        'Genetic-sex_Female_0_0': 8,
        'Handedness-chiralitylaterality_Righthanded_0_0': 9,
        'Hearing-difficultyproblems_No_0_0': 10,
        'Longstanding-illness-disability-or-infirmity_No_0_0': 11,
        'Maternal-smoking-around-birth_No_0_0': 12,
        'Milk-type-used_Semiskimmed_0_0': 13,
        'Mineral-and-other-dietary-supplements_None-of-the-above_0_0': 14,
        'Mother-still-alive_No_0_0': 15,
        'Pacemaker_No_0_0': 16,
        'Part-of-a-multiple-birth_No_0_0': 17,
        'Past-tobacco-smoking_I-have-never-smoked_0_0': 18,
        'Past-tobacco-smoking_Smoked-on-most-or-all-days_0_0': 19,
        'Pork-intake_Less-than-once-a-week_0_0': 20,
        'Pork-intake_Never_0_0': 21,
        'Poultry-intake_24-times-a-week_0_0': 22,
        'Poultry-intake_Once-a-week_0_0': 23,
        'Processed-meat-intake_Less-than-once-a-week_0_0': 24,
        'Processed-meat-intake_Once-a-week_0_0': 25,
        'Salt-added-to-food_Neverrarely_0_0': 26,
        'Salt-added-to-food_Sometimes_0_0': 27,
        'Smoking-status_Previous_0_0': 28,
        'Smokingsmokers-in-household_No_0_0': 29,
        'Types-of-physical-activity-in-last-4-weeks_Walking-for-pleasure-not-as-a-means-of-transport_0_0': 30,
        'Types-of-transport-used-excluding-work_Carmotor-vehicle_0_0': 31,
        'Vitamin-and-mineral-supplements_None-of-the-above_0_0': 32,
        'Wears-glasses-or-contact-lenses_Yes_0_0': 33,
        'Wheeze-or-whistling-in-the-chest-in-last-year_No_0_0': 35,
    },
)

categorical_phenotypes_78 = TensorMap(
    'categorical-phenotypes-78', Interpretation.CATEGORICAL, path_prefix='categorical', annotation_units=64,
    channel_map={
        'Adopted-as-a-child_No_0_0': 0,
        'Alcohol-intake-versus-10-years-previously_Less-nowadays_0_0': 1,
        'Alcohol-usually-taken-with-meals_It-varies_0_0': 2,
        'Alcohol-usually-taken-with-meals_Yes_0_0': 3,
        'Beef-intake_Less-than-once-a-week_0_0': 4,
        'Bread-type_Wholemeal-or-wholegrain_0_0': 5,
        'Breastfed-as-a-baby_Yes_0_0': 6,
        'Cereal-type_Other-eg-Cornflakes-Frosties_0_0': 7,
        'Cheese-intake_24-times-a-week_0_0': 8,
        'Coffee-type_Instant-coffee_0_0': 9,
        'Comparative-body-size-at-age-10_About-average_0_0': 10,
        'Comparative-body-size-at-age-10_Thinner_0_0': 11,
        'Comparative-height-size-at-age-10_About-average_0_0': 12,
        'Country-of-birth-UKelsewhere_England_0_0': 13,
        'Current-tobacco-smoking_No_0_0': 14,
        'Drive-faster-than-motorway-speed-limit_Neverrarely_0_0': 15,
        'Duration-walking-for-pleasure_Between-30-minutes-and-1-hour_0_0': 16,
        'Ease-of-skin-tanning_Get-moderately-tanned_0_0': 17,
        'FI1-numeric-addition-test_15_0_0': 18,
        'FI3-word-interpolation_Adult_0_0': 19,
        'FI4-positional-arithmetic_6_0_0': 20,
        'FI7-synonym_Cease_0_0': 21,
        'Facial-ageing_Younger-than-you-are_0_0': 21,
        'Father-still-alive_No_0_0': 22,
        'Father-still-alive_Yes_0_0': 23,
        'Fracturedbroken-bones-in-last-5-years_No_0_0': 24,
        'Frequency-of-stair-climbing-in-last-4-weeks_610-times-a-day_0_0': 25,
        'Genetic-sex_Female_0_0': 26,
        'Had-menopause_Yes_0_0': 27,
        'Hair-colour-natural-before-greying_Dark-brown_0_0': 28,
        'Hair-colour-natural-before-greying_Light-brown_0_0': 29,
        'Handedness-chiralitylaterality_Righthanded_0_0': 30,
        'Hearing-difficultyproblems_No_0_0': 31,
        'Hot-drink-temperature_Hot_0_0': 32,
        'Hot-drink-temperature_Very-hot_0_0': 33,
        'Lambmutton-intake_Less-than-once-a-week_0_0': 34,
        'Lambmutton-intake_Never_0_0': 35,
        'Major-dietary-changes-in-the-last-5-years_No_0_0': 36,
        'Major-dietary-changes-in-the-last-5-years_Yes-because-of-other-reasons_0_0': 37,
        'Maternal-smoking-around-birth_No_0_0': 38,
        'Milk-type-used_Semiskimmed_0_0': 39,
        'Mineral-and-other-dietary-supplements_None-of-the-above_0_0': 40,
        'Mother-still-alive_No_0_0': 41,
        'Mother-still-alive_Yes_0_0': 42,
        'Mouthteeth-dental-problems_None-of-the-above_0_0': 43,
        'Nonoily-fish-intake_Once-a-week_0_0': 44,
        'Oily-fish-intake_Less-than-once-a-week_0_0': 45,
        'Oily-fish-intake_Once-a-week_0_0': 46,
        'Pain-types-experienced-in-last-month_Headache_0_0': 47,
        'Pain-types-experienced-in-last-month_None-of-the-above_0_0': 48,
        'Part-of-a-multiple-birth_No_0_0': 49,
        'Past-tobacco-smoking_I-have-never-smoked_0_0': 50,
        'Past-tobacco-smoking_Smoked-on-most-or-all-days_0_0': 51,
        'Pork-intake_Less-than-once-a-week_0_0': 52,
        'Pork-intake_Never_0_0': 53,
        'Poultry-intake_24-times-a-week_0_0': 54,
        'Poultry-intake_Once-a-week_0_0': 55,
        'Processed-meat-intake_Less-than-once-a-week_0_0': 56,
        'Processed-meat-intake_Once-a-week_0_0': 57,
        'Salt-added-to-food_Neverrarely_0_0': 58,
        'Salt-added-to-food_Sometimes_0_0': 59,
        'Shortness-of-breath-walking-on-level-ground_No_0_0': 60,
        'Skin-colour_Fair_0_0': 61,
        'Smoking-status_Never_0_0': 62,
        'Smoking-status_Previous_0_0': 63,
        'Smokingsmokers-in-household_No_0_0': 64,
        'Spread-type_Other-type-of-spreadmargarine_0_0': 65,
        'Types-of-physical-activity-in-last-4-weeks_Other-exercises-eg-swimming-cycling-keep-fit-bowling_0_1': 66,
        'Types-of-physical-activity-in-last-4-weeks_Walking-for-pleasure-not-as-a-means-of-transport_0_0': 67,
        'Types-of-transport-used-excluding-work_Carmotor-vehicle_0_0': 68,
        'Types-of-transport-used-excluding-work_Walk_0_1': 69,
        'Usual-walking-pace_Brisk-pace_0_0': 70,
        'Usual-walking-pace_Steady-average-pace_0_0': 71,
        'Variation-in-diet_Sometimes_0_0': 72,
        'Vitamin-and-mineral-supplements_None-of-the-above_0_0': 73,
        'Wears-glasses-or-contact-lenses_Yes_0_0': 74,
        'Weight-change-compared-with-1-year-ago_No-weigh-about-the-same_0_0': 75,
        'Weight-change-compared-with-1-year-ago_Yes-gained-weight_0_0': 76,
        'Wheeze-or-whistling-in-the-chest-in-last-year_No_0_0': 77,
    },
)

categorical_phenotypes_134 = TensorMap(
        'categorical-phenotypes-134', Interpretation.CATEGORICAL, path_prefix='categorical', annotation_units=64,
        channel_map={
            'Adopted-as-a-child_No_0_0': 0, 'Adopted-as-a-child_No_2_0': 1,
            'Alcohol-intake-frequency_Once-or-twice-a-week_0_0': 2,
            'Alcohol-intake-versus-10-years-previously_About-the-same_0_0': 3,
            'Alcohol-intake-versus-10-years-previously_Less-nowadays_2_0': 4,
            'Alcohol-usually-taken-with-meals_Yes_2_0': 5, 'Beef-intake_Less-than-once-a-week_0_0': 6,
            'Beef-intake_Less-than-once-a-week_2_0': 7,
            'Bread-type_Wholemeal-or-wholegrain_0_0': 10,
            'Bread-type_Wholemeal-or-wholegrain_2_0': 11,
            'Breastfed-as-a-baby_Yes_0_0': 12, 'Breathing-problems-during-period-of-job_No_0_0': 13,
            'Breathing-problems-during-period-of-job_No_0_1': 14,
            'Cereal-type_Oat-cereal-eg-Ready-Brek-porridge_0_0': 17,
            'Cheese-intake_24-times-a-week_0_0': 18, 'Cheese-intake_24-times-a-week_2_0': 19,
            'Coffee-type_Instant-coffee_0_0': 20, 'Comparative-body-size-at-age-10_Thinner_0_0': 21,
            'Comparative-height-size-at-age-10_About-average_0_0': 22,
            'Country-of-birth-UKelsewhere_England_0_0': 23, 'Country-of-birth-UKelsewhere_England_2_0': 24,
            'Current-tobacco-smoking_No_0_0': 25,
            'Current-tobacco-smoking_No_2_0': 26,
            'Drive-faster-than-motorway-speed-limit_Neverrarely_2_0': 29,
            'Drive-faster-than-motorway-speed-limit_Sometimes_0_0': 30,
            'Ease-of-skin-tanning_Get-moderately-tanned_0_0': 31,
            'Ever-had-breast-cancer-screening-mammogram_Yes_0_0': 32,
            'Ever-had-breast-cancer-screening-mammogram_Yes_2_0': 33,
            'Ever-had-cervical-smear-test_Yes_0_0': 34, 'Ever-had-cervical-smear-test_Yes_2_0': 35,
            'Ever-taken-oral-contraceptive-pill_Yes_0_0': 36,
            'Ever-taken-oral-contraceptive-pill_Yes_2_0': 37,
            'Ever-used-hormonereplacement-therapy-HRT_No_0_0': 38,
            'Eye-problemsdisorders_None-of-the-above_0_0': 39,
            'Eye-problemsdisorders_None-of-the-above_2_0': 40, 'FI1-numeric-addition-test_15_2_0': 41,
            'FI3-word-interpolation_Adult_2_0': 42, 'FI4-positional-arithmetic_6_2_0': 43,
            'FI6-conditional-arithmetic_69_2_0': 44, 'FI7-synonym_Cease_2_0': 45,
            'Facial-ageing_Younger-than-you-are_0_0': 46, 'Facial-ageing_Younger-than-you-are_2_0': 47,
            'Father-still-alive_No_0_0': 48, 'Father-still-alive_No_2_0': 49,
            'Father-still-alive_Yes_0_0': 50, 'Fracturedbroken-bones-in-last-5-years_No_0_0': 51,
            'Fracturedbroken-bones-in-last-5-years_No_2_0': 52, 'Genetic-sex_Female_0_0': 53,
            'Had-menopause_Yes_2_0': 54, 'Hair-colour-natural-before-greying_Dark-brown_0_0': 55,
            'Handedness-chiralitylaterality_Righthanded_0_0': 56,
            'Handedness-chiralitylaterality_Righthanded_2_0': 57, 'Hearing-difficultyproblems_No_0_0': 58,
            'Hearing-difficultyproblems_No_2_0': 59, 'Hot-drink-temperature_Hot_0_0': 60,
            'Hot-drink-temperature_Hot_2_0': 61, 'Lambmutton-intake_Less-than-once-a-week_0_0': 62,
            'Lambmutton-intake_Less-than-once-a-week_2_0': 63,
            'Longstanding-illness-disability-or-infirmity_No_0_0': 64,
            'Longstanding-illness-disability-or-infirmity_No_2_0': 65,
            'Major-dietary-changes-in-the-last-5-years_No_0_0': 66,
            'Major-dietary-changes-in-the-last-5-years_No_2_0': 67,
            'Maternal-smoking-around-birth_No_0_0': 68,
            'Milk-type-used_Semiskimmed_0_0': 73,
            'Milk-type-used_Semiskimmed_2_0': 74,
            'Milk-type-used_Skimmed_0_0': 75,
            'Mineral-and-other-dietary-supplements_None-of-the-above_0_0': 76,
            'Mineral-and-other-dietary-supplements_None-of-the-above_2_0': 77,
            'Mother-still-alive_No_2_0': 78,
            'Mother-still-alive_Yes_0_0': 79,
            'Mouthteeth-dental-problems_None-of-the-above_0_0': 80,
            'Mouthteeth-dental-problems_None-of-the-above_2_0': 81, 'Noisy-workplace_No_2_0': 82,
            'Nonoily-fish-intake_Once-a-week_0_0': 83, 'Nonoily-fish-intake_Once-a-week_2_0': 84,
            'Oily-fish-intake_Less-than-once-a-week_0_0': 85,
            'Overall-health-rating_Good_0_0': 88,
            'Overall-health-rating_Good_2_0': 89,
            'Pain-types-experienced-in-last-month_None-of-the-above_0_0': 92,
            'Part-of-a-multiple-birth_No_0_0': 93, 'Part-of-a-multiple-birth_No_2_0': 94,
            'Past-tobacco-smoking_I-have-never-smoked_0_0': 95,
            'Past-tobacco-smoking_I-have-never-smoked_2_0': 96, 'Pork-intake_Less-than-once-a-week_0_0': 97,
            'Pork-intake_Less-than-once-a-week_2_0': 98, 'Poultry-intake_24-times-a-week_0_0': 99,
            'Poultry-intake_24-times-a-week_2_0': 100, 'Processed-meat-intake_Less-than-once-a-week_2_0': 101,
            'Salt-added-to-food_Neverrarely_0_0': 102, 'Salt-added-to-food_Neverrarely_2_0': 103,
            'Shortness-of-breath-walking-on-level-ground_No_2_0': 104, 'Skin-colour_Fair_0_0': 105,
            'Smoking-status_Never_0_0': 106, 'Smoking-status_Never_2_0': 107,
            'Smokingsmokers-in-household_No_0_0': 108, 'Smokingsmokers-in-household_No_2_0': 109,
            'Spread-type_Butterspreadable-butter_2_0': 110,
            'Spread-type_Other-type-of-spreadmargarine_0_0': 111,
            'Types-of-physical-activity-in-last-4-weeks_Other-exercises-eg-swimming-cycling-keep-fit-bowling_0_1': 112,
            'Types-of-physical-activity-in-last-4-weeks_Other-exercises-eg-swimming-cycling-keep-fit-bowling_2_1': 113,
            'Types-of-physical-activity-in-last-4-weeks_Walking-for-pleasure-not-as-a-means-of-transport_0_0': 114,
            'Types-of-physical-activity-in-last-4-weeks_Walking-for-pleasure-not-as-a-means-of-transport_2_0': 115,
            'Types-of-transport-used-excluding-work_Carmotor-vehicle_0_0': 116,
            'Types-of-transport-used-excluding-work_Carmotor-vehicle_2_0': 117,
            'Types-of-transport-used-excluding-work_Walk_0_1': 118,
            'Types-of-transport-used-excluding-work_Walk_2_1': 119,
            'UK-Biobank-assessment-centre_Cheadle-imaging_2_0': 120, 'Usual-walking-pace_Brisk-pace_0_0': 121,
            'Usual-walking-pace_Brisk-pace_2_0': 122, 'Usual-walking-pace_Steady-average-pace_0_0': 123,
            'Usual-walking-pace_Steady-average-pace_2_0': 124, 'Variation-in-diet_Sometimes_0_0': 125,
            'Variation-in-diet_Sometimes_2_0': 126,
            'Vitamin-and-mineral-supplements_None-of-the-above_0_0': 129,
            'Vitamin-and-mineral-supplements_None-of-the-above_2_0': 130,
            'Wears-glasses-or-contact-lenses_Yes_0_0': 131, 'Wears-glasses-or-contact-lenses_Yes_2_0': 132,
            'Weight-change-compared-with-1-year-ago_No-weigh-about-the-same_0_0': 133,
            'Weight-change-compared-with-1-year-ago_No-weigh-about-the-same_2_0': 8,
            'Wheeze-or-whistling-in-the-chest-in-last-year_No_0_0': 9,
            'Wheeze-or-whistling-in-the-chest-in-last-year_No_2_0': 15,
            'Worked-with-materials-containing-asbestos_Rarelynever_0_0': 16,
            'Worked-with-materials-containing-asbestos_Rarelynever_0_1': 27,
            'Worked-with-paints-thinners-or-glues_Rarelynever_0_0': 28,
            'Worked-with-paints-thinners-or-glues_Rarelynever_0_1': 69,
            'Worked-with-pesticides_Rarelynever_0_0': 70, 'Worked-with-pesticides_Rarelynever_0_1': 71,
            'Workplace-full-of-chemical-or-other-fumes_Rarelynever_0_0': 72,
            'Workplace-full-of-chemical-or-other-fumes_Rarelynever_0_1': 86,
            'Workplace-had-a-lot-of-cigarette-smoke-from-other-people-smoking_Rarelynever_0_0': 87,
            'Workplace-had-a-lot-of-diesel-exhaust_Rarelynever_0_0': 90,
            'Workplace-had-a-lot-of-diesel-exhaust_Rarelynever_0_1': 91,
            'Workplace-very-dusty_Rarelynever_0_0': 127, 'Workplace-very-dusty_Rarelynever_0_1': 128,
        },
)
